#!/usr/bin/env python3
"""
Classes to use git as a storage repository for threat model metadata
"""
import logging
import os
from pathlib import Path
from io import StringIO
import sh
from sh.contrib import git
from sh import rm
import utils.shell as shell
import utils.match as match
from utils import load_yaml
from utils.error import StorageError

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

class GitStorage:

    # Class variable to indicate if we are in a containerised environment
    containerised = False

    def __init__(self, config:dict, execution_env):

        self.gitrepo_config = config.get("gitrepo", {}) 
        self.base_storage_dir = self.gitrepo_config.get("base-storage-dir", "/tmp/")
        self.remote = self.gitrepo_config.get("remote")
        if match.is_empty(self.remote):
            logger.error("Remote repo is empty")
            raise StorageError("storage.no-git-repo", None)
        self.default_branch = self.gitrepo_config.get("default-branch", "approved")
        self.is_entered = False

        self.repodirname = "tm_mgmt"
        self.repodir = str(Path(self.base_storage_dir).joinpath(self.repodirname))

        # Need to setup git depending on the environment we are in
        execution_env.setupGit(self.remote, self.base_storage_dir)


    def _not_entered(self):
        if not self.is_entered:
            logger.error(f"Cannot perform action outside of 'with' statement")
            raise StorageError("internal-error", {})

        return False


    def __enter__(self):

        buf = StringIO()
        
        if Path(self.repodir).is_dir():
            logger.warning(f"Directory '{self.repodir}' already exists")

        if not shell.run(self.base_storage_dir, git.clone, ["--no-checkout", self.remote, self.repodirname], _out=buf, _err_to_out=True):
            outdata = buf.getvalue()
            logger.debug(f"{outdata}")
            logger.error(f"Could not clone repo '{self.remote}'")
            raise StorageError("internal-error", {})

        #if not shell.run(self.repodir, git, ["sparse-checkout", "init", "--cone"]):
        if not shell.run(self.repodir, git, ["sparse-checkout", "init"]):
            logger.error(f"Could not sparse-checkout init")
            raise StorageError("internal-error", {})

        # Above command will set sparse-checkout pattern, but to retrieve root dir contents, we need to do a checkout (git > 2.25)
        if not shell.run(self.repodir, git, ["checkout"]):
            # This could have failed because there is no remote branch to checkout i.e. repo is brand new
            if not self.remote_branch_exists(self.default_branch):
                # Create local branch
                if not shell.run(self.repodir, git.checkout, ["-b", self.default_branch]):
                    logger.error(f"Unable to create local branch '{self.default_branch}'")
                    raise StorageError("internal-error", {})    
                # Do an initial local commit
                if not shell.run(self.repodir, git.commit, ["--allow-empty", "-m", "Creating default branch"]):
                    logger.error(f"Unable to commit empty branch '{self.default_branch}'")
                    raise StorageError("internal-error", {})
                # Create default remote branch
                if not shell.run(self.repodir, git.push, ["-u", "origin", self.default_branch + ":" + self.default_branch]):
                    logger.error(f"Unable to push initial commit for branch '{self.default_branch}'")
                    raise StorageError("internal-error", {})
            else:
                logger.error(f"Could not checkout root directory contents")
                raise StorageError("internal-error", {})
        

        self.is_entered = True

        return self


    def checkout_directory(self, dir_name:str):

        # Fetch the directory containing the metadata for the TM ID.  If it doesn't exist, nothing is downloaded
        if not shell.run(self.repodir, git, ["sparse-checkout", "add", dir_name]):
            logger.error(f"Could not sparse-checkout directory '{dir_name}'")
            raise StorageError("internal-error", {})


    def remote_branch_exists(self, branch:str):

        remote_branch_exists_var = False

        def _check_for_remote_branch(output):
            nonlocal remote_branch_exists_var
            if not remote_branch_exists_var:
                for line in str(output).splitlines():
                    if line.endswith("refs/heads/" + self.branch):
                        remote_branch_exists_var = True
                        return

        if not shell.run(os.getcwd(), git, ["ls-remote", "--heads", self.remote], _out=_check_for_remote_branch):
            logger.error(f"Could not get list of remote heads for '{self.remote}'")
            raise StorageError("internal-error", {})
        
        return remote_branch_exists_var

    def _is_default_branch(self, branch:str):
        if match.is_empty(branch) or match.equals(self.default_branch, branch):
            logger.error(f"Cannot commit directly to default branch")
            raise StorageError("internal-error", {})

        return False

    def branch_update(self, branch:str):
        """ Switches to either a local branch using existing remote branch (fetching any existing changes on that branch), or creates a new local branch """

        if self._not_entered():
            return

        if self._is_default_branch(branch):
            return

        self.branch = branch

        if self.remote_branch_exists(branch):
            if not shell.run(self.repodir, git.checkout, ["-b", self.branch, "origin/" + self.branch]):
                logger.error(f"Could not checkout remote branch '{self.branch}'")
                raise StorageError("internal-error", {})
        else:
            if not shell.run(self.repodir, git.checkout, ["-b", self.branch]):
                logger.error(f"Could not checkout local branch '{self.branch}'")
                raise StorageError("internal-error", {})
    
    def branch_replace(self, branch:str):
        """ Switches to a local branch, deleting any existing remote branch (ignoring any existing changes on that branch) """

        if self._not_entered():
            return

        if self._is_default_branch(branch):
            return

        self.branch = branch

        if self.remote_branch_exists(branch):
            if not shell.run(self.repodir, git.push, ["origin", "-d", self.branch]):
                logger.error(f"Failed to delete remote branch '{self.branch}'")
                raise StorageError("internal-error", {})

        if not shell.run(self.repodir, git.checkout, ["-b", self.branch]):
            logger.error(f"Failed to checkout local branch '{self.branch}'")
            raise StorageError("internal-error", {})

    def is_merged_to_default(self, branch:str):

        if self._not_entered():
            return

        buf = StringIO()

        # i.e. git branch -r --merged approved
        if not shell.run(self.repodir, git.branch, ["-r", "--merged", self.default_branch], _out=buf, _err_to_out=True):
            logger.error(f"Could not determine if '{self.branch}' has been merged or not")
            raise StorageError("internal-error", {})

        outdata = buf.getvalue()
        logger.debug(f"{outdata}")

        for line in outdata.splitlines():
            if line.endswith(branch):
                return True

        return False


    def commit(self, message:str = ""):

        if self._not_entered():
            return

        if self._is_default_branch(self.branch):
            return

        if not shell.run(self.repodir, git.add, ["."]):
            logger.error(f"Failed to git add local changes")
            raise StorageError("internal-error", {})
        if not shell.run(self.repodir, git.commit, ["-m", message]):
            logger.error(f"Failed to git commit local changes")
            raise StorageError("internal-error", {})
        if not shell.run(self.repodir, git.push, ["-u", "origin", self.branch]):
            logger.error(f"Failed to git push local changes to remote")
            raise StorageError("internal-error", {})
    
    def __exit__(self, exc_type, exc_value, traceback):

        # Delete the repo we checked out.  
        shell.run(self.base_storage_dir, rm, ["-rf", self.repodir])

        self.is_entered = False

    def load_yaml(self, class_list:list, relative_file_path:str) -> dict:

        file_path = Path(self.repodir).joinpath(relative_file_path)
        if file_path.is_file():
            return load_yaml.yaml_file_to_dict(file_path, class_list)

        return None

    def write_yaml(self, relative_file_path:str, contents):

        file_path = Path(self.repodir).joinpath(relative_file_path)
        if not file_path.is_file():
            # Try to make the directory, as that at least needs to exist
            file_path.parent.mkdir(parents=True, exist_ok=True)

        load_yaml.class_to_yaml_file(contents, file_path)




#!/usr/bin/env python3
"""
Utility methods for models (dict->list->dict structures)
"""

import logging

import utils.logging
logger = logging.getLogger(utils.logging.getLoggerName(__name__))

def recurse(model, key_value_callback, list_entry_callback):
    """
    Recurses through a model calling callback functions on every dict key/value and list entry

    Returns: None, if the callbacks don't terminate recursion early, otherwise it returns the callback return value that terminated recursion
    """

    def _stub_callback(*args):
        return True, None

    if key_value_callback is None:
        key_value_callback = _stub_callback
    if list_entry_callback is None:
        list_entry_callback = _stub_callback

    def _internal_recurse(model, value):
        # The return value of this function is only changed if teh recursion exits early, otherwise the input 'value' is returned

        if isinstance(model, dict):
            # We have found a row of data
            for dict_key, dict_value in model.items():
                if key_value_callback is not None:
                    # The purpose of this callback is to do something to every entry in a row or determine something about the row
                    keep_going, return_value = key_value_callback(dict_key, dict_value, value)
                    if not keep_going:
                        return keep_going, return_value
                            
                # If an entry of a row is itself a complex structure we recurse into it
                if isinstance(dict_value, dict) or isinstance(dict_value, list):
                    keep_going, loop_return_value = _internal_recurse(dict_value, return_value)
                    if not keep_going:
                        return keep_going, loop_return_value

        if isinstance(model, list):
            # We have found a table
            for list_entry in model:
                if isinstance(list_entry, dict) or isinstance(list_entry, list):
                    # The purpose of this callback is do something with an entire row
                    keep_going, return_value = list_entry_callback(list_entry, value)
                    if not keep_going:
                        return keep_going, return_value
                    keep_going, loop_return_value = _internal_recurse(list_entry, return_value)
                    if not keep_going:
                        return keep_going, loop_return_value
        
        return True, value
    
    _, return_value =_internal_recurse(model, None)

    return return_value


# In order to find the location of certain tagged data we need to tag keys with the rows they are in
def assign_row_identifiers(model):

    def _key_value_callback(dict_key, dict_value, rowIDKey):
        # We dont know where in the hierarchy we are, but for dicts we just add the rowIDKey
        if rowIDKey is not None:
            dict_key.addProperty("rowID", rowIDKey)

        return True, rowIDKey

    def _list_entry_callback(list_entry, rowIDKey):
        if isinstance(list_entry, dict):
            local_rowIDKey = rowIDKey
            # Rows are stored as dict
            for dict_key, dict_value in list_entry.items():
                if dict_key.hasTag("row-identifier"):
                    #print(f"Found key tagged with 'row-identifier - {dict_key}")
                    # We found a "row-identifier" for an entry in this row, let's use that from now on
                    local_rowIDKey = dict_key
                    # We need to report errors with the value for the "row-identiifer" key, so let's store the value as a property of the key
                    dict_key.addProperty("value", dict_value)
                    # We need to be able to get sibling data in a row, so also store a reference to the whole row against the 'row-identifier' key
                    dict_key.addProperty("row", list_entry)

                    # There should only be 1 row-identifier per row
                    break
            # Let's assign the the rowIDKey to all elements of this row (the 'row-identifier' field might not have been the first)
            # This will include child elements as well
            return True, local_rowIDKey
        else:
            # Not a row so recurse using existing rowIDKey
            return True, rowIDKey

    return recurse(model, _key_value_callback, _list_entry_callback)


# In order to find the location of certain tagged data we need to tag keys with their parent keys
def assign_parent_keys(model):

    def _key_value_callback(dict_key, dict_value, parentKey):
        # For every key, assign it a parentKey
        dict_key.addProperty("parentKey", parentKey)
        # If we recurse further, then this dict_key becomes the parentKey for descendants
        return True, dict_key

    def _list_entry_callback(list_entry, parentKey):
        # Descendants of a list entry share the same parentKey, so just pass it along
        return True, parentKey
        
    return recurse(model, _key_value_callback, _list_entry_callback)
    